/*
 * driver for PWM (Pulse Width Modulator) for TPU controller
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * Copyright 2011 Atmark Techno, Inc. All Rights Reserved.
 */

#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/err.h>
#include <linux/pwm.h>
#include <linux/clk.h>
#include <linux/io.h>
#include <linux/slab.h>

#include "tpu-pwm.h"

#define TPU_CH0		0x10
#define TPU_CH1		0x50
#define TPU_CH2		0x90
#define TPU_CH3		0xd0
#define TPU_CH_OFFSET	0x40

#define TPU_TSTR	0x00

#define TPU_TCRn	0x00
#define TPU_TMDRn	0x04
#define TPU_TIORn	0x08
#define TPU_TIERn	0x0c
#define TPU_TSRn	0x10
#define TPU_TCNTn	0x14
#define TPU_TGRnA	0x18
#define TPU_TGRnB	0x1c
#define TPU_TGRnC	0x20
#define TPU_TGRnD	0x24

struct pwm_device {
	struct list_head list;
	struct platform_device *pdev;

	void __iomem *base;
	struct clk *clk;

	int use_count;
	const char *label;
};

static DEFINE_MUTEX(pwm_lock);
static LIST_HEAD(pwm_list);

static void pwm_writew(struct pwm_device *pwm, u16 val, u32 reg)
{
	__raw_writew(val, (pwm->base + TPU_CH0 +
			   TPU_CH_OFFSET * pwm->pdev->id + reg));
}

int pwm_enable(struct pwm_device *pwm)
{
	u16 val;

	val = __raw_readw(pwm->base + TPU_TSTR);
	val |= 1 << pwm->pdev->id;
	__raw_writew(val, pwm->base + TPU_TSTR);

	return 0;
}
EXPORT_SYMBOL(pwm_enable);

void pwm_disable(struct pwm_device *pwm)
{
	struct platform_device *pdev = pwm->pdev;
	struct tpu_pwm_platform_data *pdata = pdev->dev.platform_data;
	u16 val;

	if (pdata->is_low_at_disable) {
		pwm_writew(pwm, 0x0000, TPU_TGRnA);
	} else {
		val = __raw_readw(pwm->base + TPU_TSTR);
		val &= ~(1 << pwm->pdev->id);
		__raw_writew(val, pwm->base + TPU_TSTR);
	}
}
EXPORT_SYMBOL(pwm_disable);

int pwm_config(struct pwm_device *pwm, int duty_ns, int period_ns)
{
	u32 duty, period, clk_rate;

	clk_rate = clk_get_rate(pwm->clk);

	period = clk_rate / (NSEC_PER_SEC / period_ns);
	if (period > 0xffff)
		return -EINVAL;

	if (duty_ns) {
		duty = clk_rate / (NSEC_PER_SEC / duty_ns) + 1;
		if (duty > 0xffff)
			return -EINVAL;
	} else {
		duty = 0;
	}

	pwm_writew(pwm, (u16)duty, TPU_TGRnA);
	pwm_writew(pwm, (u16)period, TPU_TGRnB);

	return 0;
}
EXPORT_SYMBOL(pwm_config);

struct pwm_device *pwm_request(int pwm_id, const char *label)
{
	struct pwm_device *pwm;

	mutex_lock(&pwm_lock);
	list_for_each_entry(pwm, &pwm_list, list) {
		if (pwm->pdev->id == pwm_id) {
			if (pwm->use_count == 0) {
				pwm->use_count++;
				pwm->label = label;
				pwm_writew(pwm, 0x0002, TPU_TMDRn);
				pwm_writew(pwm, 0x0040, TPU_TCRn);
				pwm_writew(pwm, 0x0002, TPU_TIORn);
				mutex_unlock(&pwm_lock);
				return pwm;
			} else {
				mutex_unlock(&pwm_lock);
				return ERR_PTR(-EBUSY);
			}
		}
	}
	mutex_unlock(&pwm_lock);
	return ERR_PTR(-ENODEV);
}
EXPORT_SYMBOL(pwm_request);

void pwm_free(struct pwm_device *pwm)
{
	pwm_disable(pwm);
	mutex_lock(&pwm_lock);
	pwm->use_count = 0;
	pwm->label = NULL;
	mutex_unlock(&pwm_lock);
}
EXPORT_SYMBOL(pwm_free);

static int __devinit tpu_pwm_probe(struct platform_device *pdev)
{
	struct pwm_device *pwm;
	struct clk *clk = clk_get(NULL, "tpu");
	struct resource *res;
	struct tpu_pwm_platform_data *pdata = pdev->dev.platform_data;
	int ret;

	pwm = kzalloc(sizeof(struct pwm_device), GFP_KERNEL);
	if (!pwm) {
		dev_err(&pdev->dev, "failed to allocate memory\n");
		return -ENOMEM;
	}

	pwm->clk = clk_get(&pdev->dev, "tpu");
	if (IS_ERR(pwm->clk)) {
		ret = PTR_ERR(pwm->clk);
		goto pwm_free;
	}

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!res) {
		ret = -ENODEV;
		goto pwm_free;
	}

	pwm->base = ioremap_nocache(res->start, resource_size(res));
	if (!pwm->base) {
		ret = -ENXIO;
		dev_err(&pdev->dev, "unable to ioremap TPU registers.\n");
		goto clk_release;
	}

	mutex_lock(&pwm_lock);
	list_add_tail(&pwm->list, &pwm_list);
	mutex_unlock(&pwm_lock);

	pwm->pdev = pdev;
	platform_set_drvdata(pdev, pwm);

	if (pdata->init)
		pdata->init(pwm);

	clk_enable(clk);

	return 0;

clk_release:
	clk_put(pwm->clk);
pwm_free:
	kfree(pwm);
	return ret;
}

static int __devexit tpu_pwm_remove(struct platform_device *pdev)
{
	struct pwm_device *pwm;

	pwm = platform_get_drvdata(pdev);
	iounmap(pwm->base);
	clk_put(pwm->clk);
	kfree(pwm);

	return 0;
}

static struct platform_driver tpu_pwm_driver = {
	.driver		= {
		.name	= "tpu_pwm",
	},
	.probe		= tpu_pwm_probe,
	.remove		= __devexit_p(tpu_pwm_remove),
};

static int __init tpu_pwm_init(void)
{
	return platform_driver_register(&tpu_pwm_driver);
}
arch_initcall(tpu_pwm_init);

static void __exit tpu_pwm_exit(void)
{
	platform_driver_unregister(&tpu_pwm_driver);
}
module_exit(tpu_pwm_exit);

MODULE_LICENSE("GPL v2");
MODULE_AUTHOR("Atmark Techno, Inc.");
