#include <linux/kernel.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/io.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/fb.h>
#include <linux/clk.h>

#include <mach/irqs.h>
#include <mach/rmobile_sdenc.h>

#ifdef DEBUG
#define d_dbg(fmt, args...) { if (1) printk(fmt, args); 0; }
#else
#define d_dbg(fmt, args...) { if (0) printk(fmt, args); 0; }
#endif

#ifdef TRACE_DEBUG
#define trace() { if (1) printk("*** %s()\n", __func__); 0; }
#else
#define trace() { if (0) printk("*** %s()\n", __func__); 0; }
#endif

struct sh_mobile_sdenc_priv {
	u32 *base;

	struct notifier_block nb;
	int blanked;
};

struct sdenc_mode {
	u32 rate;
	u32 xres;
	u32 yres;
	u32 refresh;
	u32 vmode;
};

static struct sh_mobile_sdenc_priv *sdenc_priv;
static struct sdenc_mode sdenc_modes[] = {
	/*  720x480i@60 */  {  27000000,  720,  480, 120, FB_VMODE_INTERLACED },
};
static u32 current_sdenc = -1;
static const char *source_dev_name;

static struct work_struct sdenc_work;

static DECLARE_MUTEX(sdenc_sem);

static u32 find_sdenc_mode(struct fb_videomode *mode)
{
	int i;
	for (i=0; i<ARRAY_SIZE(sdenc_modes); i++) {
		if (sdenc_modes[i].xres == mode->xres &&
		    sdenc_modes[i].yres == mode->yres &&
		    sdenc_modes[i].refresh == mode->refresh &&
		    sdenc_modes[i].vmode == mode->vmode)
			return i;
	}

	return -1;
}

static void sdenc_write(u32 offset, u16 val)
{
	struct sh_mobile_sdenc_priv *priv = sdenc_priv;

	iowrite16(val, (u32)priv->base | offset);
}

static void setup_sdenc(void)
{
	trace();

	down(&sdenc_sem);

	/* setup SDENC */
	sdenc_write(0x0300, 0x0f3f);
	sdenc_write(0x0302, 0x00ff);
	sdenc_write(0x0308, 0x0000);
	sdenc_write(0x00c0, 0x3000);
	sdenc_write(0x00c2, 0xff00);
	sdenc_write(0x00c4, 0xc900);
	sdenc_write(0x00c6, 0x21f0);
	sdenc_write(0x00c8, 0x7c1f);
	sdenc_write(0x00ca, 0x0000);
	sdenc_write(0x00fc, 0x0000);
	sdenc_write(0x00cc, 0x15db);
	sdenc_write(0x00fe, 0x00db);
	sdenc_write(0x00ce, 0x0000);
	sdenc_write(0x00d0, 0x0000);
	sdenc_write(0x00d2, 0x0000);
	sdenc_write(0x0100, 0x0000);
	sdenc_write(0x0102, 0x0000);
	sdenc_write(0x00d4, 0x0000);
	sdenc_write(0x0104, 0x0000);
	sdenc_write(0x00d6, 0x0000);
	sdenc_write(0x00d8, 0x0000);
	sdenc_write(0x00da, 0x0080);
	sdenc_write(0x00dc, 0x0000);
	sdenc_write(0x00de, 0x0040);
	sdenc_write(0x00f8, 0x4565);
	sdenc_write(0x00fa, 0x003d);
	sdenc_write(0x0108, 0x0000);
	sdenc_write(0x010a, 0x0000);
	sdenc_write(0x010c, 0x0000);
	sdenc_write(0x010e, 0x0000);

	up(&sdenc_sem);
}

static void sh_mobile_sdenc_setup(struct work_struct *work)
{
	struct sh_mobile_sdenc_priv *priv = sdenc_priv;

	trace();

	if (priv->blanked)
		return;

	if (current_sdenc != -1)
		setup_sdenc();
}

static int sh_mobile_sdenc_event_handler(struct notifier_block *nb,
					unsigned long action, void *data)
{
	struct sh_mobile_sdenc_priv *priv = sdenc_priv;
	struct fb_event *event = data;
	struct fb_info *info = event->info;
	struct fb_videomode *mode;

	trace();

	if (strncmp(source_dev_name, dev_name(info->device),
		    strlen(source_dev_name)))
		return 0;

	switch (action) {
	case FB_EVENT_MODE_CHANGE:
		mode = event->data;
		current_sdenc = find_sdenc_mode(mode);

		if (current_sdenc != -1)
			schedule_work(&sdenc_work);
		else
			sdenc_write(0x0300, 0x0fff);
		break;

	case FB_EVENT_BLANK:
		if (*(int *)event->data)
			priv->blanked = 1; /* blank */
		else
			priv->blanked = 0; /* unblank */

		if (priv->blanked || current_sdenc == -1)
			sdenc_write(0x0300, 0x0fff);
		else
			sdenc_write(0x0300, 0x0f3f);
		break;
	default:
		break;
	}
	return 0;
}

static int __devinit sh_mobile_sdenc_probe(struct platform_device *pdev)
{
	struct sh_mobile_sdenc_info *pdata = pdev->dev.platform_data;
	struct sh_mobile_sdenc_priv *priv;
	struct resource *res;

	trace();

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!res) {
		dev_err(&pdev->dev, "cannot get platform resources\n");
		return -ENOENT;
	}

	sdenc_priv = kzalloc(sizeof(struct sh_mobile_sdenc_priv), GFP_KERNEL);
	if (!sdenc_priv) {
		dev_err(&pdev->dev, "unable to allocate memory\n");
		return -ENOMEM;
	}

	priv = sdenc_priv;
	priv->base = ioremap_nocache(res->start, (res->end - res->start) + 1);

	source_dev_name = pdata->source_dev_name;
	INIT_WORK(&sdenc_work, sh_mobile_sdenc_setup);

	priv->nb.notifier_call = sh_mobile_sdenc_event_handler;
	fb_register_client(&priv->nb);

	return 0;
}

static int sh_mobile_sdenc_remove(struct platform_device *pdev)
{
	struct sh_mobile_sdenc_priv *priv = sdenc_priv;

	trace();

	fb_unregister_client(&priv->nb);
	cancel_work_sync(&sdenc_work);
	iounmap(priv->base);
	kfree(sdenc_priv);

	return 0;
}

static struct platform_driver sh_mobile_sdenc_driver = {
	.driver = {
		.name = "sh_mobile_sdenc",
		.owner = THIS_MODULE,
	},
	.probe = sh_mobile_sdenc_probe,
	.remove = sh_mobile_sdenc_remove,
};

static int __init rmobile_sdenc_init(void)
{
	trace();
	return platform_driver_register(&sh_mobile_sdenc_driver);
}

static void __exit rmobile_sdenc_exit(void)
{
	trace();
	platform_driver_unregister(&sh_mobile_sdenc_driver);
}

module_init(rmobile_sdenc_init);
module_exit(rmobile_sdenc_exit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("R-Mobile A1 SDENC driver");
