/*
 *  File Name       : RSPF0_spu_api.c
 *  Function        : SPU OPEN SOUND I/F
 *  Release Version : Ver 1.00
 *  Release Date    : 2011/11/30
 *
 *  Copyright (C) 2011 Renesas Electronics Corporation
 *
 *  This program is free software;you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by Free
 *  Softwere Foundation; either version 2 of License, or (at your option) any
 *  later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY;
 *  without even the implied warrnty of MERCHANTABILITY or FITNESS FOR A
 *  PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program;
 *  If not, write to the Free Software Foundation, Inc., 51 Franklin St,
 *  Fifth Floor, Boston, MA  02110-1301  USA
 *
 */
#include "RSPF0_spu_driver.h"
#include "RSPF0_spu_api.h"

#define API_RECEIVE_HANDLER_NUM  (16)

#define SPUAPI_GET_CLASS_ID(x)   (((x) >> 12) & 0xff)

enum {
	SPUAPI_SYNC_STATE_INIT = 0,
	SPUAPI_SYNC_STATE_WAIT,
	SPUAPI_SYNC_STATE_END,
};

static SPUAPI_Func spuapi_func[SPUAPI_DSP_MAX];
static unsigned long spuapi_address_mode;

static void spuapi_receive_callback0(unsigned long api_id, unsigned long *data,
	unsigned long size);
static void spuapi_receive_callback1(unsigned long api_id, unsigned long *data,
	unsigned long size);

static unsigned long sync_api_id[SPUAPI_DSP_MAX];
static unsigned long sync_state[SPUAPI_DSP_MAX];

typedef struct {
	unsigned long class_id;
	void (*hdl)(unsigned long dsp_id, unsigned long api,
				unsigned long *data, unsigned long size);
} tSPUAPIReceiveHandler;

static tSPUAPIReceiveHandler receive_handler[SPUAPI_DSP_MAX]
	[API_RECEIVE_HANDLER_NUM];

static void SPUAPI_ResponseCallback(unsigned long dsp_num, unsigned long api,
	unsigned long *data, unsigned long size);

/*-------------------------------------------------------
Name    : SPUAPI_Init()
Outline : Initialize receive_handler for CPU.
Input   : dsp_id      : DSP ID
Output  : error code (0)
-------------------------------------------------------*/
long SPUAPI_Init(unsigned long dsp_id)
{
	int i;
	for (i = 0; i < API_RECEIVE_HANDLER_NUM; i++) {
		receive_handler[dsp_id][i].hdl = 0;
		receive_handler[dsp_id][i].class_id = 0;
	}
	return 0;
}

/*-------------------------------------------------------
Name    : SPUAPI_Open()
Outline : Set user function and download Firmware
Input   : dsp_id       : DSP ID
	: *func        : Pointers of user function
	: *data        : Pointers of Firmware of SPU2
	: size         : Sizes of Firmware of SPU2
	: capacity     : Sizes of Banks of SPU2
	: address_mode : Address mode
Output  : error code
-------------------------------------------------------*/
long SPUAPI_Open(unsigned long dsp_id,
			SPUAPI_Func *func,
			const unsigned long *data[3],
			const unsigned long size[3],
			unsigned long capacity[3],
			unsigned long address_mode
			) {
	unsigned long param[8];
	unsigned long i;
	long result;
	SPUDRV_STRUCT(UserFunc) drv_func;

	/* parameter check */
	if (!data || !func || !func->reg_read || !func->reg_write ||
	!func->set_event || !func->wait_for_event ||
	!func->enter_critical_section || !func->leave_critical_section ||
	address_mode > SPUAPI_ADDRESS_MODE_32BIT)
		return -1;

	/* boot parameter - X/Y memory capacity */
	for (i = 0; i < 2; i++)
		param[i] = capacity[i+1];
	/* boot parameter - reserve */
	for (i = 2; i < 8; i++)
		param[i] = 0;

	drv_func.set_event              = func->set_event;
	drv_func.wait_for_event         = func->wait_for_event;
	drv_func.enter_critical_section = func->enter_critical_section;
	drv_func.leave_critical_section = func->leave_critical_section;
	drv_func.read                   = func->reg_read;
	drv_func.write                  = func->reg_write;

	result = -1;
	switch (dsp_id) {
	case SPUAPI_DSP0:
		result = SPUDRV_FUNC(Open)(dsp_id, spuapi_receive_callback0,
				&drv_func, 1, address_mode, data, size, param);
		break;
	case SPUAPI_DSP1:
		result = SPUDRV_FUNC(Open)(dsp_id, spuapi_receive_callback1,
				&drv_func, 1, address_mode, data, size, param);
		break;
	/*default:*/
		/*break;*/
	}
	if (result != 0)
		return -3;

	spuapi_func[dsp_id].reg_read       = func->reg_read;
	spuapi_func[dsp_id].reg_write      = func->reg_write;
	spuapi_func[dsp_id].set_event      = func->set_event;
	spuapi_func[dsp_id].wait_for_event = func->wait_for_event;
	spuapi_func[dsp_id].enter_critical_section =
						func->enter_critical_section;
	spuapi_func[dsp_id].leave_critical_section =
						func->leave_critical_section;

	spuapi_address_mode = address_mode;

	sync_api_id[dsp_id] = 0;
	sync_state[dsp_id] = SPUAPI_SYNC_STATE_INIT;

	return 0;
}

/*-------------------------------------------------------
Name    : SPUAPI_RegisterApiReceiveHandler()
Outline : Set handler and this class id
Input   : dsp_id       : DSP ID
	: class_id     : CLASS ID
	: *hdl         : Pointer of handler function
Output  : error code
-------------------------------------------------------*/
long SPUAPI_RegisterApiReceiveHandler(unsigned long dsp_id,
		unsigned long class_id, void (*hdl)(unsigned long dsp_id,
		unsigned long api, unsigned long *data, unsigned long size)) {
	unsigned long i;

	for (i = 0; i < API_RECEIVE_HANDLER_NUM; i++) {
		if (receive_handler[dsp_id][i].hdl == 0)
			break;
	}
	if (i == API_RECEIVE_HANDLER_NUM)
		return -1;
	receive_handler[dsp_id][i].hdl = hdl;
	receive_handler[dsp_id][i].class_id = class_id;

	return 0;
}

/*-------------------------------------------------------
Name    : spuapi_receive_callback0()
Outline : Receive function from DSP0 of SPU2
Input   : api_id : API ID from DSP0 of SPU2
	: *data  : Pointer of Parameter from DSP0 of SPU2
	: size   : size of data
Output  : none
-------------------------------------------------------*/
static void spuapi_receive_callback0(
					unsigned long api_id,
					unsigned long *data,
					unsigned long size) {
	SPUAPI_ResponseCallback(0, api_id, data, size);
}

/*-------------------------------------------------------
Name    : spuapi_receive_callback1()
Outline : Receive function from DSP1 of SPU2
Input   : api_id : API ID from DSP1 of SPU2
	: *data  : Pointer of Parameter from DSP1 of SPU2
	: size   : size of data
Output  : none
-------------------------------------------------------*/
static void spuapi_receive_callback1(
					unsigned long api_id,
					unsigned long *data,
					unsigned long size) {
	SPUAPI_ResponseCallback(1, api_id, data, size);
}

/*-------------------------------------------------------
Name    : SPUAPI_ResponseCallback()
Outline : Excute Receive handler Function from SPU2
Input   : dsp_num : DSP ID
	: api     : API ID
	: data    : Pointer of Parameter of SPU2
	: size    : size of data
Output  : none
-------------------------------------------------------*/
static void SPUAPI_ResponseCallback(unsigned long dsp_num, unsigned long api,
				unsigned long *data, unsigned long size) {
	unsigned long i;
	unsigned long class_id;

	class_id = SPUAPI_GET_CLASS_ID(api);

	/* API callback is immediately executed. */
	for (i = 0; i < API_RECEIVE_HANDLER_NUM; i++) {
		if (receive_handler[dsp_num][i].class_id == class_id) {
			receive_handler[dsp_num][i].hdl
						(dsp_num, api, data, size);
			/* Check the release of the synchronized state */
			if (sync_state[dsp_num] == SPUAPI_SYNC_STATE_WAIT) {
				if (api == sync_api_id[dsp_num]) {
					sync_state[dsp_num] =
							SPUAPI_SYNC_STATE_END;
					sync_api_id[dsp_num] = 0;
					spuapi_func[dsp_num].set_event();
				}
			}
			break;
		}
	}
}

/*-------------------------------------------------------
Name    : SPUAPI_SendApi()
Outline : Excute Request function each component to SPU2
Input   : dsp_num : DSP ID
	: api     : API ID
	: data    : Pointer of Parameter to SPU2
	: size    : size of data
	: sync_id : Synchronization / Asynchronization
Output  : error code
-------------------------------------------------------*/
long SPUAPI_SendApi(
			unsigned long dsp_id,
			unsigned long api,
			unsigned long *data,
			unsigned long size,
			unsigned long sync_id)
{
	if (sync_id != 0) {
		sync_api_id[dsp_id] = sync_id;
		sync_state[dsp_id]  = SPUAPI_SYNC_STATE_WAIT;
	}

	if (SPUDRV_FUNC(SendApi)(dsp_id, api, data, size) != 0) {
		if (sync_state[dsp_id] == SPUAPI_SYNC_STATE_WAIT)
			sync_state[dsp_id] = SPUAPI_SYNC_STATE_INIT;
		return -1;
	}

	if (sync_id != 0) {

		if (spuapi_func[dsp_id].wait_for_event(1000) < 0) {
			/* if timeout occurred, */
			sync_state[dsp_id] = SPUAPI_SYNC_STATE_INIT;
			return -1;
		}

		sync_state[dsp_id] = SPUAPI_SYNC_STATE_INIT;
	}

	return 0;
}

/*-------------------------------------------------------
Name    : SPUAPI_Interrupt()
Outline : Excute Interrupt function
Input   : dsp_num : DSP ID
Output  : none
-------------------------------------------------------*/
void SPUAPI_Interrupt(unsigned long dsp_id)
{
	SPUDRV_FUNC(ISR)(dsp_id);
}

/*-------------------------------------------------------
Name    : SPUAPI_DmaAddress()
Outline : Check DMA Address
Input   : addr : Address for transfer
Output  : error code
-------------------------------------------------------*/
unsigned long SPUAPI_DmaAddress(unsigned long addr)
{
	if (spuapi_address_mode == SPUAPI_ADDRESS_MODE_29BIT)
		return addr & 0x1fffffff;
	else
		return addr;
}

/*-------------------------------------------------------
Name    : SPUAPI_CheckDmaParam()
Outline : Check SHway restriction
Input   : start_addr : Start Address
	: word       : Size of transfer
	: tr_mode    : setting of unit
Output  : error code
-------------------------------------------------------*/
long SPUAPI_CheckDmaParam(unsigned long start_addr, unsigned long word,
							unsigned long tr_mode) {
	if (word) {
		if (tr_mode >= 4) {
			if (start_addr & 0x1f)
				return -1;
			word = ((word + 8 - 1) / 8) * 8;
		}

		if ((start_addr / 0x800000) !=
			((start_addr + word - 1) / 0x800000))
			return -1;
	}
	return 0;
}

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("SPU OPEN SOUND I/F Ver 1.0");
MODULE_AUTHOR("Renesas Electronics");
