/*
 *  File Name       : RSPF0_spu_api_fmsiport.c
 *  Function        : SPU OPEN SOUND I/F
 *  Release Version : Ver 1.00
 *  Release Date    : 2011/11/30
 *
 *  Copyright (C) 2011 Renesas Electronics Corporation
 *
 *  This program is free software;you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by Free
 *  Softwere Foundation; either version 2 of License, or (at your option) any
 *  later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY;
 *  without even the implied warrnty of MERCHANTABILITY or FITNESS FOR A
 *  PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program;
 *  If not, write to the Free Software Foundation, Inc., 51 Franklin St,
 *  Fifth Floor, Boston, MA  02110-1301  USA
 *
 */
#include "RSPF0_spu_api.h"
#include "RSPF0_spu_api_fmsiport.h"

#define CLASS_ID 0x11

/* sign extention macro */
#define INT24_TO_INT(x)             \
	(((x) & 0x00800000) ?       \
	((x) | 0xff000000) :  (x))

enum {
	API_FMSIPORT_CREATE_REQUEST             = SPUAPI_API_ID(CLASS_ID, 0x00),
	API_FMSIPORT_CREATE_RESPONSE            = SPUAPI_API_ID(CLASS_ID, 0x80),
	API_FMSIPORT_DELETE_REQUEST             = SPUAPI_API_ID(CLASS_ID, 0x01),
	API_FMSIPORT_DELETE_RESPONSE            = SPUAPI_API_ID(CLASS_ID, 0x81),
	API_FMSIPORT_SET_PARAM_REQUEST          = SPUAPI_API_ID(CLASS_ID, 0x02),
	API_FMSIPORT_SET_PARAM_RESPONSE         = SPUAPI_API_ID(CLASS_ID, 0x82),
	API_FMSIPORT_ALLOCATE_REQUEST           = SPUAPI_API_ID(CLASS_ID, 0x03),
	API_FMSIPORT_ALLOCATE_RESPONSE          = SPUAPI_API_ID(CLASS_ID, 0x83),
	API_FMSIPORT_ENABLE_REQUEST             = SPUAPI_API_ID(CLASS_ID, 0x04),
	API_FMSIPORT_ENABLE_RESPONSE            = SPUAPI_API_ID(CLASS_ID, 0x84),
	API_FMSIPORT_DISABLE_REQUEST            = SPUAPI_API_ID(CLASS_ID, 0x05),
	API_FMSIPORT_DISABLE_RESPONSE           = SPUAPI_API_ID(CLASS_ID, 0x85),
	API_FMSIPORT_RESET_ERROR_REQUEST        = SPUAPI_API_ID(CLASS_ID, 0x06),
	API_FMSIPORT_RESET_ERROR_RESPONSE       = SPUAPI_API_ID(CLASS_ID, 0x86),
	API_FMSIPORT_ERROR                      = SPUAPI_API_ID(CLASS_ID, 0x87),
	API_FMSIPORT_GET_SAMPLE_COUNT_REQUEST   = SPUAPI_API_ID(CLASS_ID, 0x08),
	API_FMSIPORT_GET_SAMPLE_COUNT_RESPONSE  = SPUAPI_API_ID(CLASS_ID, 0x88),
	API_FMSIPORT_CLEAR_SAMPLE_COUNT_REQUEST = SPUAPI_API_ID(CLASS_ID, 0x09),
	API_FMSIPORT_CLEAR_SAMPLE_COUNT_RESPONSE =
						SPUAPI_API_ID(CLASS_ID, 0x89),
	API_FMSIPORT_FLUSH_REQUEST              = SPUAPI_API_ID(CLASS_ID, 0x0A),
	API_FMSIPORT_FLUSH_RESPONSE             = SPUAPI_API_ID(CLASS_ID, 0x8A),
	API_FMSIPORT_SET_FADEOUT_REQUEST        = SPUAPI_API_ID(CLASS_ID, 0x0B),
	API_FMSIPORT_SET_FADEOUT_RESPONSE       = SPUAPI_API_ID(CLASS_ID, 0x8B),
};

#define FMSIPORT_MAX_NUM 8

#define FMSIPORT_PROCESSED_SIZE_PARAM_MASK  0x00FFFFFF

typedef struct {
	sSPUFmsiPort *port;
	tSPUFmsiPortErrorCallBack error_callback;
} sFmsiPortHandle;

static sFmsiPortHandle handle[SPUAPI_DSP_MAX][FMSIPORT_MAX_NUM];
static unsigned long response[SPUAPI_DSP_MAX][SPUAPI_MESSAGE_LENGTH];

const unsigned long fsi_bitsel[4] = {	0, /* PORT A OUT */
					4, /* PORT A IN  */
					8, /* PORT B OUT */
					12 /* PORT B IN  */
};
const unsigned long fmsi_bitsel[4] = {	0,  /* PORT S OUT */
					4,  /* PORT S IN  */
					12, /* RESERVE    */
					8   /* PORT P IN  */
};

long SPUFmsiPort_Allocate(unsigned long dsp_id, unsigned long io_addr,
	unsigned long isr, unsigned long *fmsi_id, unsigned long bit_select) {
	unsigned long api_id;
	unsigned long size;
	unsigned long response_api;
	long result;
	unsigned long data[6];
	const unsigned long *bitsel;
	unsigned long i;

	if (dsp_id >= SPUAPI_DSP_MAX || io_addr >= 0x10000 || isr >
	SPUAPI_ISR2 || !fmsi_id || bit_select >= SPUAPI_FMSIPORT_BITSELECT_MAX)
		return SPUAPI_ERROR_PARAM;

	if (bit_select == SPUAPI_FMSIPORT_BITSELECT_FMSI)
		bitsel = fmsi_bitsel;
	else
		bitsel = fsi_bitsel;

	api_id = API_FMSIPORT_ALLOCATE_REQUEST;
	size = 6;
	data[0] = io_addr;
	data[1] = isr;
	for (i = 0; i < 4; i++)
		data[2+i] = bitsel[i];

	response_api = API_FMSIPORT_ALLOCATE_RESPONSE;

	if (SPUAPI_SendApi(dsp_id, api_id, data, size, response_api) < 0)
		return SPUAPI_ERROR_DRIVER;

	result = (long)response[dsp_id][0];
	if (result == 0)
		*fmsi_id = response[dsp_id][1];
	return result;
}

long SPUFmsiPort_Create(unsigned long dsp_id, unsigned long fmsi_id,
				unsigned long port_id, sSPUFmsiPort *port) {
	unsigned long api_id;
	unsigned long size;
	unsigned long response_api;
	long result;
	unsigned long i;
	unsigned long data[1];

	if (!port || dsp_id >= SPUAPI_DSP_MAX || port_id >= SPUFMSIPORT_MAX)
		return SPUAPI_ERROR_PARAM;

	for (i = 0; i < FMSIPORT_MAX_NUM; i++) {
		if (!handle[dsp_id][i].port)
			break;
	}
	if (i == FMSIPORT_MAX_NUM)
		return SPUAPI_ERROR_RESOURCE;

	api_id = API_FMSIPORT_CREATE_REQUEST;
	size = 1;
	data[0] = (fmsi_id << 2) + port_id;
	response_api = API_FMSIPORT_CREATE_RESPONSE;

	if (SPUAPI_SendApi(dsp_id, api_id, data, size, response_api) < 0)
		return SPUAPI_ERROR_DRIVER;

	result = (long)response[dsp_id][0];
	if (result == 0) {
		port->port_id = response[dsp_id][1];
		port->dsp_id = dsp_id;
		handle[dsp_id][i].port = port;
	}

	return result;
}

long SPUFmsiPort_SetCallBack(sSPUFmsiPort *port, tSPUFmsiPortErrorCallBack cb)
{
	unsigned long i;

	if (!port)
		return SPUAPI_ERROR_PARAM;
	for (i = 0; i < FMSIPORT_MAX_NUM; i++) {
		if (handle[port->dsp_id][i].port) {
			if (handle[port->dsp_id][i].port->port_id ==
			port->port_id)
				break;
		}
	}
	if (i == FMSIPORT_MAX_NUM)
		return SPUAPI_ERROR_NO_OBJECT;
	handle[port->dsp_id][i].error_callback = cb;
	return 0;
}

long SPUFmsiPort_Delete(sSPUFmsiPort *port)
{
	unsigned long api_id;
	unsigned long size;
	unsigned long response_api;
	long result;
	unsigned long data[1];
	unsigned long i;

	if (!port)
		return SPUAPI_ERROR_PARAM;

	api_id       = API_FMSIPORT_DELETE_REQUEST;
	size         = 1;
	data[0]      = port->port_id;
	response_api = API_FMSIPORT_DELETE_RESPONSE;

	if (SPUAPI_SendApi(port->dsp_id, api_id, data, size, response_api) < 0)
		return SPUAPI_ERROR_DRIVER;

	result = (long)response[port->dsp_id][0];
	if (result == 0) {
		port->port_id = 0;

		for (i = 0; i < FMSIPORT_MAX_NUM; i++) {
			if (handle[port->dsp_id][i].port == port) {
				handle[port->dsp_id][i].port = 0;
				break;
			}
		}
		if (i == FMSIPORT_MAX_NUM)
			return SPUAPI_ERROR_NO_OBJECT;
	}

	return result;
}

static long default_request(unsigned long dsp_id, unsigned long api_id,
	unsigned long *data, unsigned long size, unsigned long response_api,
							unsigned long sync) {
	if (sync > SPUAPI_WAIT_RESPONSE)
		return SPUAPI_ERROR_PARAM;

	if (sync == SPUAPI_NO_RESPONSE) {
		response_api = 0;
		api_id = SPUAPI_NO_RESPONSE_API(api_id);
	}

	if (SPUAPI_SendApi(dsp_id, api_id, data, size, response_api) < 0)
		return SPUAPI_ERROR_DRIVER;

	if (sync == SPUAPI_NO_RESPONSE)
		return 0;
	else
		return response[dsp_id][0];
}

static const unsigned long fmsiport_ch_max_sample[] = {
					256, 128, 64, 64, 32, 32, 32, 32 };
long SPUFmsiPort_SetParam(sSPUFmsiPort *port, sSPUBuffer *buffer,
				sSPUFmsiPortParam *param, unsigned long sync) {
	unsigned long api_id;
	unsigned long size;
	unsigned long response_api;
	unsigned long data[6];

	if (!port || !buffer || !param || param->ch > 8 || param->ch == 0 ||
	param->threshold_mode >= SPUAPI_FMSIPORT_THRESHOLD_MODE_MAX)
		return SPUAPI_ERROR_PARAM;
	if (param->threshold_mode == SPUAPI_FMSIPORT_THRESHOLD_MODE_SAMPLE) {
		if (param->threshold_sample == 0 || param->threshold_sample >=
		fmsiport_ch_max_sample[param->ch - 1])
			return SPUAPI_ERROR_PARAM;
	}

	if (param->output_start_threshold * param->ch > buffer->size)
		return SPUAPI_ERROR_PARAM;

	api_id  = API_FMSIPORT_SET_PARAM_REQUEST;
	size    = 6;
	data[0] = port->port_id;
	data[1] = buffer->buffer_id;
	data[2] = param->ch;
	data[3] = (param->threshold_mode << 16) |
			(param->threshold_sample & 0xffff);
	data[4] = 0;
	data[5] = param->output_start_threshold;
	response_api = API_FMSIPORT_SET_PARAM_RESPONSE;

	return default_request(port->dsp_id, api_id, data, size, response_api,
									sync);
}

long SPUFmsiPort_Enable(sSPUFmsiPort *port, unsigned long sync)
{
	unsigned long api_id;
	unsigned long size;
	unsigned long response_api;
	unsigned long data[1];

	if (!port)
		return SPUAPI_ERROR_PARAM;

	api_id       = API_FMSIPORT_ENABLE_REQUEST;
	size         = 1;
	data[0]      = port->port_id;
	response_api = API_FMSIPORT_ENABLE_RESPONSE;

	return default_request(port->dsp_id, api_id, data, size, response_api,
									sync);
}

long SPUFmsiPort_Disable(sSPUFmsiPort *port, unsigned long sync)
{
	unsigned long api_id;
	unsigned long size;
	unsigned long response_api;
	unsigned long data[1];

	if (!port)
		return SPUAPI_ERROR_PARAM;

	api_id       = API_FMSIPORT_DISABLE_REQUEST;
	size         = 1;
	data[0]      = port->port_id;
	response_api = API_FMSIPORT_DISABLE_RESPONSE;

	return default_request(port->dsp_id, api_id, data, size, response_api,
									sync);
}

long SPUFmsiPort_Flush(sSPUFmsiPort *port, unsigned long sync)
{
	unsigned long api_id;
	unsigned long size;
	unsigned long response_api;
	unsigned long data[1];

	if (!port)
		return SPUAPI_ERROR_PARAM;

	api_id       = API_FMSIPORT_FLUSH_REQUEST;
	size         = 1;
	data[0]      = port->port_id;
	response_api = API_FMSIPORT_FLUSH_RESPONSE;

	return default_request(port->dsp_id, api_id, data, size, response_api,
									sync);
}

long SPUFmsiPort_ResetError(sSPUFmsiPort *port, unsigned long sync)
{
/*long SPUFmsiPort_ResetError(sSPUFmsiPort *port, unsigned long reset_type,
							unsigned long sync) {*/
	unsigned long api_id;
	unsigned long size;
	unsigned long response_api;
	unsigned long data[1];

/*	if (!port || reset_type > SPUAPI_FMSIPORT_RESET_ALL)*/
	if (!port)
		return SPUAPI_ERROR_PARAM;

	api_id       = API_FMSIPORT_RESET_ERROR_REQUEST;
	size         = 1;
	data[0]      = port->port_id;
/*	data[1]      = reset_type;*/
	response_api = API_FMSIPORT_RESET_ERROR_RESPONSE;

	return default_request(port->dsp_id, api_id, data, size, response_api,
									sync);
}

long SPUFmsiPort_GetSampleCount(sSPUFmsiPort *port, unsigned long *count)
{
	unsigned long api_id;
	unsigned long size;
	unsigned long response_api;
	unsigned long data[1];
	long result;

	if (!port)
		return SPUAPI_ERROR_PARAM;

	api_id       = API_FMSIPORT_GET_SAMPLE_COUNT_REQUEST;
	size         = 1;
	data[0]      = port->port_id;
	response_api = API_FMSIPORT_GET_SAMPLE_COUNT_RESPONSE;

	result = default_request(port->dsp_id, api_id, data, size, response_api,
							SPUAPI_WAIT_RESPONSE);

	if (!result) {
		if (count)
			*count = ((FMSIPORT_PROCESSED_SIZE_PARAM_MASK &
				response[port->dsp_id][2]) << 24) +
				(FMSIPORT_PROCESSED_SIZE_PARAM_MASK &
				response[port->dsp_id][3]);
	}

	return result;
}

long SPUFmsiPort_ClearSampleCount(sSPUFmsiPort *port, unsigned long sync)
{
	unsigned long api_id;
	unsigned long size;
	unsigned long response_api;
	unsigned long data[1];

	if (!port)
		return SPUAPI_ERROR_PARAM;

	api_id       = API_FMSIPORT_CLEAR_SAMPLE_COUNT_REQUEST;
	size         = 1;
	data[0]      = port->port_id;
	response_api = API_FMSIPORT_CLEAR_SAMPLE_COUNT_RESPONSE;

	return default_request(port->dsp_id, api_id, data, size, response_api,
									sync);
}

static sFmsiPortHandle *get_handle_from_port_id(unsigned long dsp_id,
	unsigned long port_id) {
	unsigned long i;

	for (i = 0; i < FMSIPORT_MAX_NUM; i++) {
		if (handle[dsp_id][i].port && handle[dsp_id][i].port->port_id ==
		port_id)
			return &handle[dsp_id][i];
	}
	return 0;
}

static void SPUFmsiPort_ApiReceiveHandler(unsigned long dsp_id,
	unsigned long api_id, unsigned long *data, unsigned long size) {
	unsigned long i;
	sFmsiPortHandle *handle;

	switch (api_id) {
	case API_FMSIPORT_ERROR:
		handle = get_handle_from_port_id(dsp_id, data[0]);
		if (handle && handle->error_callback)
			handle->error_callback(handle->port, data[1], data[2]);
		break;
	default:
		for (i = 0; i < size; i++)
			response[dsp_id][i] = INT24_TO_INT(data[i]);
								/* pgr0788 */
		break;
	}
}

long SPUFmsiPort_Init(unsigned long dsp_id)
{
	unsigned long i;

	for (i = 0; i < FMSIPORT_MAX_NUM; i++)
		handle[dsp_id][i].port = 0;

	if (SPUAPI_RegisterApiReceiveHandler(dsp_id, CLASS_ID,
	SPUFmsiPort_ApiReceiveHandler) != 0)
		return -1;
	return 0;
}

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("SPU OPEN SOUND I/F Ver 1.0");
MODULE_AUTHOR("Renesas Electronics");
