/*
 *  File Name       : RSPF0_spu_api_soundv2.c
 *  Function        : SPU OPEN SOUND I/F
 *  Release Version : Ver 1.00
 *  Release Date    : 2011/11/30
 *
 *  Copyright (C) 2011 Renesas Electronics Corporation
 *
 *  This program is free software;you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by Free
 *  Softwere Foundation; either version 2 of License, or (at your option) any
 *  later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY;
 *  without even the implied warrnty of MERCHANTABILITY or FITNESS FOR A
 *  PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program;
 *  If not, write to the Free Software Foundation, Inc., 51 Franklin St,
 *  Fifth Floor, Boston, MA  02110-1301  USA
 *
 */
#include "RSPF0_spu_api_soundv2.h"

#define CLASS_ID 0x42

/* sign extention macro */
#define INT24_TO_INT(x)             \
	(((x) & 0x00800000) ?       \
	((x) | 0xff000000) :  (x))

/* API Parameter Checker */
#define SPUSOUND_SRC_OV_MAX				0x00200000
#define SPUSOUND_SRC_OV_MIN				0x00008000
#define SPUSOUND_PROCESSED_SIZE_PARAM_MAX	0x000001
#define SPUSOUND_PROCESSED_SIZE_PARAM_MASK  0x00FFFFFF
#define SOUND_MAX_NUM	(3)			/* MAX Number of SOUND Comp */

/* API ID for SOUND Comp */
enum {
	API_SOUND_CREATE_REQUEST	= SPUAPI_API_ID(CLASS_ID, 0x000000),
	API_SOUND_CREATE_RESPONSE	= SPUAPI_API_ID(CLASS_ID, 0x000080),
	API_SOUND_DELETE_REQUEST	= SPUAPI_API_ID(CLASS_ID, 0x000001),
	API_SOUND_DELETE_RESPONSE	= SPUAPI_API_ID(CLASS_ID, 0x000081),
	API_SOUND_SET_PARAM_REQUEST	= SPUAPI_API_ID(CLASS_ID, 0x000002),
	API_SOUND_SET_PARAM_RESPONSE	= SPUAPI_API_ID(CLASS_ID, 0x000082),
	API_SOUND_GET_PARAM_REQUEST	= SPUAPI_API_ID(CLASS_ID, 0x000003),
	API_SOUND_GET_PARAM_RESPONSE	= SPUAPI_API_ID(CLASS_ID, 0x000083),
	API_SOUND_ENABLE_REQUEST	= SPUAPI_API_ID(CLASS_ID, 0x000004),
	API_SOUND_ENABLE_RESPONSE	= SPUAPI_API_ID(CLASS_ID, 0x000084),
	API_SOUND_DISABLE_REQUEST	= SPUAPI_API_ID(CLASS_ID, 0x000005),
	API_SOUND_DISABLE_RESPONSE	= SPUAPI_API_ID(CLASS_ID, 0x000085),
	API_SOUND_SET_SRC_REQUEST	= SPUAPI_API_ID(CLASS_ID, 0x000006),
	API_SOUND_SET_SRC_RESPONSE	= SPUAPI_API_ID(CLASS_ID, 0x000086),
	API_SOUND_CLEAR_REQUEST		= SPUAPI_API_ID(CLASS_ID, 0x00000F),
	API_SOUND_CLEAR_RESPONSE	= SPUAPI_API_ID(CLASS_ID, 0x00008F),
	API_SOUND_PROC_SIZE_REQUEST	= SPUAPI_API_ID(CLASS_ID, 0x000011),
	API_SOUND_PROC_SIZE_RESPONSE	= SPUAPI_API_ID(CLASS_ID, 0x000091)
};

/* Notify handler of SOUND Comp(ALC, VOL) */
typedef struct {
  sSPUSound *sound;
} sSoundHandle;

/* static function */
static sSoundHandle handle[SPUAPI_DSP_MAX][SOUND_MAX_NUM];
static unsigned long response[SPUAPI_DSP_MAX][SPUAPI_MESSAGE_LENGTH];


/*-----------------------------------------------------------
Name    : default_request()
Outline : Request API function for SOUND Comp
Input   : dsp_id        : Requested DSP Number
	: api_id        : Requested API function ID
	: *data         : Start ptr of parameter array
	: size          : Number of parameter
	: response_api  : Array of responsed parameter
	: sync          : Synchronization / Asynchronization
Output  : error code
-----------------------------------------------------------*/
static int default_request(unsigned long dsp_id, unsigned long api_id,
	unsigned long *data, unsigned long size, unsigned long response_api,
							unsigned long sync) {
	if (sync > SPUAPI_WAIT_RESPONSE)
		return SPUAPI_ERROR_PARAM;

	if (sync == SPUAPI_NO_RESPONSE) {
		response_api = 0;
		api_id = SPUAPI_NO_RESPONSE_API(api_id);
	}

	if (SPUAPI_SendApi(dsp_id, api_id, data, size, response_api) < 0)
		return SPUAPI_ERROR_DRIVER;

	if (sync == SPUAPI_NO_RESPONSE)
		return 0;
	else
		return response[dsp_id][0];
}



/*-----------------------------------------------------------
Name    : SPUSound_Create()
Outline : Request Create API function for SOUND Comp
Input   : dsp_id        : Requested DSP Number
	: *obj          : Ptr of SOUND Comp infomation
	: fir_mem       : Memory flag for SRC
Output  : error code
-----------------------------------------------------------*/
long SPUSound_Create(unsigned long dsp_id, sSPUSound *obj,
							unsigned long fir_mem) {
	unsigned long api_id;
	unsigned long size;
	unsigned long data[1];
	unsigned long response_api;
	long result;
	unsigned long i;

	if (!obj || dsp_id >= SPUAPI_DSP_MAX || fir_mem > FIR_EXIST)
		return SPUAPI_ERROR_PARAM;

	for (i = 0; i < SOUND_MAX_NUM; i++) {
		if (!handle[dsp_id][i].sound)
			break;
	}
	if (i == SOUND_MAX_NUM)
		return SPUAPI_ERROR_RESOURCE;

	api_id = API_SOUND_CREATE_REQUEST;
	size = 1;
	data[0] = fir_mem;
	response_api = API_SOUND_CREATE_RESPONSE;

	if (SPUAPI_SendApi(dsp_id, api_id, data, size, response_api) < 0)
		return SPUAPI_ERROR_DRIVER;

	result = (long)response[dsp_id][0];
	if (result == 0) {
		obj->obj_id = response[dsp_id][1];
		obj->dsp_id = dsp_id;
		handle[obj->dsp_id][i].sound = obj;
	}

	return result;
}

/*-----------------------------------------------------------
Name    : SPUSound_Delete()
Outline : Request Delete API function for SOUND Comp
Input   : *obj          : Ptr of SOUND Comp infomation
Output  : error code
-----------------------------------------------------------*/
long SPUSound_Delete(sSPUSound *obj)
{
	unsigned long api_id;
	unsigned long size;
	unsigned long data[1];
	unsigned long response_api;
	long result;
	unsigned long i;

	if (!obj)
		return SPUAPI_ERROR_PARAM;
	api_id = API_SOUND_DELETE_REQUEST;
	size = 1;
	data[0] = obj->obj_id;
	response_api = API_SOUND_DELETE_RESPONSE;

	if (SPUAPI_SendApi(obj->dsp_id, api_id, data, size, response_api) < 0)
		return SPUAPI_ERROR_DRIVER;
	result = (long)response[obj->dsp_id][0];
	if (result == 0) {
		obj->obj_id = 0;
		for (i = 0; i < SOUND_MAX_NUM; i++) {
			if (handle[obj->dsp_id][i].sound == obj) {
				handle[obj->dsp_id][i].sound = 0;
				break;
			}
		}
		if (i == SOUND_MAX_NUM)
			return SPUAPI_ERROR_NO_OBJECT;
	}

	return result;
}

/*-----------------------------------------------------------
Name    : SPUSound_SetParam()
Outline : Request Delete API function for SOUND Comp
Input   : *obj          : Ptr of SOUND Comp infomation
	: *buf_in       : Ptr of BUFFER for input
	: *buf_out      : Ptr of BUFFER for output
	: channel       : input chanel
	: proc_size     : Processing size
	: sync          : Synchronization / Asynchronization
Output  : error code
-----------------------------------------------------------*/
long SPUSound_SetParam(sSPUSound *obj, sSPUBuffer *buf_in, sSPUBuffer *buf_out,
	unsigned long channel, sSoundParam *param, unsigned long proc_size,
							unsigned long sync) {
	unsigned long api_id;
	unsigned long size;
	unsigned long data[10];
	unsigned long response_api;

	if (!obj || !buf_in || !buf_out || !param)
		return SPUAPI_ERROR_PARAM;
	if (channel > SOUND_STEREO)
		return SPUAPI_ERROR_PARAM;
	if (param->fir_mode > FIR_SRC)
		return SPUAPI_ERROR_PARAM;
	if ((proc_size > 0x7FFF) || (proc_size == 0))
		return SPUAPI_ERROR_PARAM;

	api_id = API_SOUND_SET_PARAM_REQUEST;
	size = 10;
	data[0] = obj->obj_id;
	data[1] = buf_in->buffer_id;
	data[2] = buf_out->buffer_id;
	data[3] = channel;
	data[4] = param->fir_mode;
	data[5] = 0;
	data[6] = 0;
	data[7] = 0;
	data[8] = 0;
	data[9] = proc_size;
	response_api = API_SOUND_SET_PARAM_RESPONSE;

	return default_request(obj->dsp_id, api_id, data, size,
		response_api, sync);
}

/*-----------------------------------------------------------
Name    : SPUSound_Enable()
Outline : Request Enable API function for SOUND Comp
Input   : *obj          : Ptr of SOUND Comp infomation
	: sync          : Synchronization / Asynchronization
Output  : error code
-----------------------------------------------------------*/
long SPUSound_Enable(sSPUSound *obj, unsigned long sync)
{
	unsigned long api_id;
	unsigned long size;
	unsigned long data[1];
	unsigned long response_api;

	if (!obj)
		return SPUAPI_ERROR_PARAM;

	api_id = API_SOUND_ENABLE_REQUEST;
	size = 1;
	data[0] = obj->obj_id;
	response_api = API_SOUND_ENABLE_RESPONSE;

	return default_request(obj->dsp_id, api_id, data, size,
		response_api, sync);
}

/*-----------------------------------------------------------
Name    : SPUSound_Disable()
Outline : Request Disable API function for SOUND Comp
Input   : *obj          : Ptr of SOUND Comp infomation
	: sync          : Synchronization / Asynchronization
Output  : error code
-----------------------------------------------------------*/
long SPUSound_Disable(sSPUSound *obj, unsigned long sync)
{
	unsigned long api_id;
	unsigned long size;
	unsigned long data[1];
	unsigned long response_api;

	if (!obj)
		return SPUAPI_ERROR_PARAM;

	api_id = API_SOUND_DISABLE_REQUEST;
	size = 1;
	data[0] = obj->obj_id;
	response_api = API_SOUND_DISABLE_RESPONSE;

	return default_request(obj->dsp_id, api_id, data, size,
		response_api, sync);
}

/*-----------------------------------------------------------
Name    : SPUSound_SetSRC()
Outline : Request SetSRC API function for SOUND Comp
Input   : *obj          : Ptr of SOUND Comp infomation
	: ov            : Ov of SRC
	: sync          : Synchronization / Asynchronization
Output  : error code
-----------------------------------------------------------*/
long SPUSound_SetSRC(sSPUSound *obj, unsigned long ov, unsigned long sync)
{
	unsigned long api_id;
	unsigned long size;
	unsigned long data[2];
	unsigned long response_api;

	if (!obj)
		return SPUAPI_ERROR_PARAM;
	if (ov > SPUSOUND_SRC_OV_MAX)
		return SPUAPI_ERROR_PARAM;
	else if (ov < SPUSOUND_SRC_OV_MIN)
		return SPUAPI_ERROR_PARAM;

	api_id = API_SOUND_SET_SRC_REQUEST;
	size = 2;
	data[0] = obj->obj_id;
	data[1] = ov;
	response_api = API_SOUND_SET_SRC_RESPONSE;

	return default_request
			(obj->dsp_id, api_id, data, size, response_api, sync);
}


/*-----------------------------------------------------------
Name    : SPUSound_Clear()
Outline : Request Clear API function for SOUND Comp
Input   : *obj          : Ptr of SOUND Comp infomation
	: sync          : Synchronization / Asynchronization
Output  : error code
-----------------------------------------------------------*/
long SPUSound_Clear(sSPUSound *obj, unsigned long sync)
{
	unsigned long api_id;
	unsigned long size;
	unsigned long data[1];
	unsigned long response_api;

	if (!obj)
		return SPUAPI_ERROR_PARAM;

	api_id = API_SOUND_CLEAR_REQUEST;
	size = 1;
	data[0] = obj->obj_id;
	response_api = API_SOUND_CLEAR_RESPONSE;

	return default_request(obj->dsp_id, api_id, data, size,
		response_api, sync);
}

/*-----------------------------------------------------------
Name    : SPUAPI_SoundReceiveHandler()
Outline : Select Response() handle function from dsp_id
		and api_id
Input   : dsp_id        : Used DSP ID
	: api_id        : resnponsed API id
	: *data         : Response Data from SPU2
Output  : error code
-----------------------------------------------------------*/
static void SPUAPI_SoundReceiveHandler(unsigned long dsp_id,
		unsigned long api_id, unsigned long *data, unsigned long size) {
	unsigned long i;

	for (i = 0; i < size; i++)
		response[dsp_id][i] = INT24_TO_INT(data[i]);	/* pgr0788 */

}

/*-----------------------------------------------------------
Name    : SPUAPI_SoundReceiveHandler()
Outline : Setting event handler and initialing handle
Input   : dsp_id        : Used DSP ID
Output  : error code
-----------------------------------------------------------*/
long SPUSound_Init(unsigned long dsp_id)
{
	unsigned long i;

	if (SPUAPI_RegisterApiReceiveHandler(dsp_id, CLASS_ID,
	SPUAPI_SoundReceiveHandler) != 0)
		return -1;

	for (i = 0; i < SOUND_MAX_NUM; i++)
		handle[dsp_id][i].sound = 0;

	return 0;
}

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("SPU OPEN SOUND I/F Ver 1.0");
MODULE_AUTHOR("Renesas Electronics");
